"""
Spatial Representations for Artificial Intelligence.

srai is a Python module integrating many geo-related algorithms in a single package with unified
API. It aims to provide simple and efficient solutions to geospatial problems that are accessible to
everybody and reusable in various contexts where geospatial data can be used.

See
https://kraina-ai.github.io/srai/
for complete documentation.
"""

__version__ = "0.7.3"
